import dateutil
from addict import Dict

from crosspm.contracts.contract import Contract
from crosspm.contracts.package_version import PackageVersion
from dohq_common.package_parsers.debian_package_name_parser import DebianPackageNameParser

PACKAGE_PROPERTY_CONTRACT_PREFFIX = 'contracts.'


class Package:

    def __init__(self, name: str, version: str, contracts: dict):
        self.name = name
        self.version = PackageVersion(version)
        self.contracts = contracts

    def __hash__(self):
        return hash(str(self.name) + str(self.version))

    def __str__(self):
        return "{}.{}({})".format(self.name, self.version, ";".join(str(s[1]) for s in sorted(self.contracts.items())))

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.name == other.name and self.version == other.version

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        return (self.name, self.version) < (other.name, other.version)

    def is_microservice(self, microservice):
        return self.name == microservice

    def has_contract(self, contract):
        return contract in self.contracts

    def has_contracts(self, contracts):
        return bool(self.contracts.keys() & contracts)

    def is_contract_lower_then(self, other):
        if other.name in self.contracts:
            return self.contracts[other.name].values != other.values

        raise BaseException("package {} has no contract {}", str(self), str(other))

    def is_any_contract_higher(self, other):
        for c in self.calc_contracts_intersection(other.contracts):
            if self.contracts[c] != other.contracts[c]:
                return True

        return False

    def calc_contracts_intersection(self, contracts):
        return self.contracts.keys() & contracts.keys()

    @staticmethod
    def create_contracts(contracts):
        return {c[0]: Contract(c[0], c[1]) for c in contracts}

    @staticmethod
    def create_contracts_from_dict(contracts):
        return {k: Contract(k, v) for k, v in contracts.items()}

    @staticmethod
    def create_package_from_tuple(package):

        if len(package) == 3:
            return Package(package[0], str(package[1]), Package.create_contracts_from_dict(parse_contracts_from_string(package[2])))

        return Package(package[0], str(package[1]), Package.create_contracts([]))

    @staticmethod
    def create_package_debian(package_name):

        p = DebianPackageNameParser.parse_from_package_name(package_name)

        return Package.create_package_from_tuple((p.package, p.fullversion))

    @staticmethod
    def create_packages(*packages):
        res = set()
        for p in packages:
            res.add(Package.create_package_from_tuple(p))

        return res


def create_artifactory_package(art_path):
    contracts = parse_contracts_from_package_properties(art_path.properties)
    debian_package = DebianPackageNameParser.parse_from_package_name(art_path.name)

    return ArtifactoryPackage(art_path, debian_package.package, debian_package.fullversion, contracts)


def parse_contracts_from_package_properties(properties):
    contracts = Dict()

    for p in properties:
        if p.startswith(PACKAGE_PROPERTY_CONTRACT_PREFFIX):
            contracts[p] = properties[p]

    return contracts


def parse_contracts_from_string(contracts_string):
    contracts = Dict()

    for cs in contracts_string.strip().split(';'):
        props = cs.strip().split('=')
        if props.__len__() == 2:
            contract, values = props
            if contract and contract.startswith(PACKAGE_PROPERTY_CONTRACT_PREFFIX):
                contracts[contract] = [v.strip() for v in values.split(',')]

    return contracts


def is_packages_contracts_graph_resolvable(packages):
    contracts = {}

    for p in packages:
        contracts_intersection = contracts.keys() & p.contracts.keys()
        for c in contracts_intersection:
            if not p.contracts[c] == contracts[c]:
                return False

        contracts.update(p.contracts)

    return True


class ArtifactoryPackage(Package):
    def __init__(self, art_path, name, version, contracts):
        super(ArtifactoryPackage, self).__init__(name, version,
                                                 Package.create_contracts_from_dict(contracts))
        self.art_path = art_path
        self.stat_pkg = None

    def strisodate_to_timestamp(self, date):
        return dateutil.parser.parse(date).timestamp()

    def pkg_stat(self):
        if not self.stat_pkg:
            self.stat = self.art_path.stat()
        return self.stat_pkg



import logging
from ordered_set import OrderedSet

from crosspm.contracts.package import is_packages_contracts_graph_resolvable
from crosspm.helpers.exceptions import CrosspmException, CROSSPM_ERRORCODE_PACKAGE_NOT_FOUND, \
    CrosspmBundleNoValidContractsGraph, CrosspmBundleTriggerPackagesHasNoValidContractsGraph, \
    CrosspmBundleTriggerPackageHidesHigherVersion


class Bundle:
    def __init__(self, deps, packages_repo, trigger_packages, enable_tp_hides_higher_version):
        # it is vital for deps to be list, orderedset (or something with insertion order savings),
        # we need the order of packages in dependencies.txt to take next package
        # when no contracts satisfied
        self._log = logging.getLogger('crosspm')

        self._deps = OrderedSet(deps)
        self._packages_repo = sorted(packages_repo, reverse=True)

        self._trigger_packages = []
        if trigger_packages:
            for tp in trigger_packages:
                self._trigger_packages.append(Bundle.find_trigger_package_in_packages_repo(tp, self._packages_repo))
                if not enable_tp_hides_higher_version:
                    validate_trigger_package_doesnt_hide_higher_version(tp, self._packages_repo)

        self._packages = dict()
        self._bundle_contracts = {}

    @staticmethod
    def find_trigger_package_in_packages_repo(trigger_package, repo_packages):

        for p in repo_packages:
            if p == trigger_package:
                return p

        raise CrosspmException(CROSSPM_ERRORCODE_PACKAGE_NOT_FOUND,
                               f"trigger_package = <{trigger_package}> NOT FOUND in repo_packages : {repo_packages}")

    def calculate(self):

        self._log.info('deps: {}'.format(self._deps))
        self._log.info('trigger_packages: {}'.format(self._trigger_packages))
        self._log.info('packages_repo: {}'.format(self._packages_repo))

        if not is_packages_contracts_graph_resolvable(self._trigger_packages):
            raise CrosspmBundleTriggerPackagesHasNoValidContractsGraph(self._trigger_packages)

        for tp in self._trigger_packages:
            self._packages[tp.name] = tp

        while True:
            rest_packages_to_find = self.rest_packages_to_find(self._deps, self._packages)
            if not rest_packages_to_find:
                break

            self.update_bundle_contracts()
            self.find_next_microservice_package(rest_packages_to_find)

        return self._packages

    def find_next_microservice_package(self, rest_packages_to_find):

        failed_contracts = set()
        microservice_packages = [i for i in self._packages_repo if i.is_microservice(rest_packages_to_find[0])]

        if not microservice_packages:
            raise CrosspmBundleNoValidContractsGraph(f"cant select next package for current contracts:\n"
                                                     f"rest_packages_to_find : {rest_packages_to_find}\n"
                                                     f"bundle.packages : {self._packages}")

        for p in [i for i in self._packages_repo if i.is_microservice(rest_packages_to_find[0])]:
            package = self.is_package_corresponds_bundle_current_contracts(failed_contracts, p, self._bundle_contracts)
            if package:
                self._package_add(package)
                return

        for package_name, package in self._packages.copy().items():
            if package.has_contracts(failed_contracts):
                if package in self._trigger_packages:
                    raise CrosspmBundleNoValidContractsGraph(
                        f"no tree resolve with trigger_packages {self._trigger_packages}, threre is no appropriate packages with specified package contracts")

                del self._packages[package_name]
                self._packages_repo.remove(package)

        return

    def select_next_microservice_package_out_of_current_contracts(self, next_packages_out_of_current_contracts,
                                                                  select_order):
        for i in select_order:
            if i in next_packages_out_of_current_contracts:
                return next_packages_out_of_current_contracts[i]

        return None

    def is_package_corresponds_bundle_current_contracts(self, failed_contracts, package,
                                                        bundle_contracts):

        intersection_package_contracts = package.calc_contracts_intersection(bundle_contracts)

        if not intersection_package_contracts:
            return package

        failed_contracts.update(intersection_package_contracts)
        for c in intersection_package_contracts:

            if package.contracts[c] == bundle_contracts[c]:
                failed_contracts.discard(c)

        if not failed_contracts:
            return package

        return None

    def remove_packages_with_higher_contracts_then(self, package_lowering_contract):

        for p in [*self._packages.values()]:
            if p.is_any_contract_higher(package_lowering_contract):
                if p in self._trigger_packages:
                    raise CrosspmBundleNoValidContractsGraph(
                        f"no tree resolve with trigger_packages {self._trigger_packages}, threre is no appropriate packages with specified package contracts")

                del self._packages[p.name]

    def update_bundle_contracts(self):
        self._bundle_contracts = dict()
        for p in self._packages.values():
            self._bundle_contracts.update(p.contracts)

    def rest_packages_to_find(self, deps, packages):
        return deps - packages.keys()

    def _package_add(self, package):
        self._packages[package.name] = package


def validate_trigger_package_doesnt_hide_higher_version(tp, packages):
    for p in [i for i in packages if i.is_microservice(tp.name)]:
        if tp.version < p.version:
            raise CrosspmBundleTriggerPackageHidesHigherVersion(tp, p)

    return True



# -*- coding: utf-8 -*-
from addict import Dict
from pathlib import PurePath

from crosspm.contracts.package_version import PackageVersion
from dohq_common.package_parsers.debian_package_name_parser import DebianPackageNameParser


class Parser2():
    def __init__(self, name, data, config):
        # super().__init__(name, data, config)
        self._name = 'repo2'
        self._rules = {}
        self._rules['path'] = [data['path']]

    def get_vars(self):
        return ['server', 'repo', 'package', 'version']

    def validate_path(self, path, params):
        p = PurePath(path)
        debian_package = DebianPackageNameParser.parse_from_package_name(p.name)
        package_version = PackageVersion(debian_package.fullversion)

        res_params = Dict()
        res_params.update(params)
        res_params.version = package_version.release

        res_params_raw = Dict()
        res_params_raw.version = debian_package.fullversion

        return True, res_params, res_params_raw

    def merge_with_mask(self, column, value):
        return value

    @staticmethod
    def split_fixed_pattern_with_file_name(path):
        """
        Split path into fixed, masked parts and filename
        :param path: e.g
https://repo.example.com/artifactory/libs-cpp-release.snapshot/boost/1.60-pm/*.*.*/vc110/x86/win/boost.*.*.*.tar.gz
        :return:
            _path_fixed: https://repo.example.com/artifactory/libs-cpp-release.snapshot/boost/1.60-pm/
            _path_pattern: *.*.*/vc100/x86/win
            _file_name_pattern: boost.*.*.*.tar.gz
        """
        _first_pattern_pos = path.find('*')
        _path_separator_pos = path.rfind('/', 0, _first_pattern_pos)
        _path_fixed = path[:_path_separator_pos]
        _path_pattern = path[_path_separator_pos + 1:]
        _file_name_pattern_separator_pos = _path_pattern.rfind('/', 0)
        _file_name_pattern = _path_pattern[_file_name_pattern_separator_pos + 1:]

        if _path_pattern.find('*') == -1 or _file_name_pattern_separator_pos == -1:
            _path_pattern = ""
        else:
            _path_pattern = _path_pattern[:_file_name_pattern_separator_pos]

        return _path_fixed, _path_pattern, _file_name_pattern



# -*- coding: utf-8 -*-
from collections import OrderedDict

import http
import logging
import os
import pathlib
import requests
import shutil
from artifactory import ArtifactoryPath

from crosspm.helpers.config import CROSSPM_DEPENDENCY_FILENAME, CROSSPM_DEPENDENCY_LOCK_FILENAME, Config  # noqa
from crosspm.helpers.exceptions import *
from crosspm.helpers.package import Package
from crosspm.helpers.parser import Parser
from dohq_common.deps_txt.deps_txt_lock_list_formatter import DepsTxtLockListFormatter


class Command(object):
    def entrypoint(self, *args, **kwargs):
        assert NotImplemented


class Downloader(Command):
    def __init__(self, config, do_load, recursive, parser_class=Parser):
        self._log = logging.getLogger('crosspm')
        self._config = config  # type: Config
        self._root_package = Package('<root>', 0, None, {self._config.name_column: '<root>'}, self, None,
                                     None)
        self.recursive = recursive

        self.do_load = do_load

        deps_path = config.deps_path.strip().strip('"').strip("'")
        self._deps_path = os.path.realpath(os.path.expanduser(deps_path))

        if not config.depslock_path:
            config.depslock_path = \
                config.deps_lock_file_name if config.deps_lock_file_name else CROSSPM_DEPENDENCY_LOCK_FILENAME
        depslock_path = config.depslock_path.strip().strip('"').strip("'")
        self._depslock_path = os.path.realpath(os.path.expanduser(depslock_path))

    def update_progress(self, msg, progress):
        self._log.info('\r{0} [{1:10}] {2}%'.format(msg, '#' * int(float(progress) / 10.0), int(progress)))

    # Get list of all packages needed to resolve all the dependencies.
    # List of Package class instances.
    def get_dependency_packages(self, package_matches, property_validate=True):

        _packages = OrderedDict()

        for i, _src in enumerate(self._config.sources()):
            self._log.info('')
            self._log.info(f'source: {_src}')

            _found_packages = _src.get_packages(self, package_matches, property_validate)

            _packages.update(
                OrderedDict([(k, v) for k, v in _found_packages.items() if _packages.get(k, None) is None]))

        return _packages

    def get_usedby_packages(self, list_or_file_path=None, property_validate=True):
        """

        :param list_or_file_path:
        :param property_validate: for `root` packages we need check property, bad if we find packages from `lock` file,
        we can skip validate part
        :return:
        """
        if list_or_file_path is None:
            list_or_file_path = self._depslock_path
            if not os.path.isfile(list_or_file_path):
                list_or_file_path = self._deps_path

        _packages = OrderedDict()
        if isinstance(list_or_file_path, str):
            self._log.info('Reading dependencies ... [%s]', list_or_file_path)
        for i, _src in enumerate(self._config.sources()):
            if i > 0:
                self._log.info('')
                self._log.info('Next source ...')
            _found_packages = _src.get_usedby(self, list_or_file_path, property_validate)
            _packages.update(
                OrderedDict([(k, v) for k, v in _found_packages.items() if _packages.get(k, None) is None]))
            if not self._config.no_fails:
                if isinstance(list_or_file_path, (list, tuple)):
                    list_or_file_path = [x for x in list_or_file_path if
                                         _packages.get(x[self._config.name_column], None) is None]
                elif isinstance(list_or_file_path, dict) and isinstance(list_or_file_path.get('raw', None), list):
                    list_or_file_path['raw'] = [x for x in list_or_file_path['raw'] if
                                                _packages.get(x[self._config.name_column], None) is None]

        return _packages

    def find_package_in_artifactory(self, package):
        return

    # Download packages or just unpack already loaded (it's up to adapter to decide)
    def download_packages(self):
        with open(self._config.depslock_path) as f:
            packages_path_in_repo = DepsTxtLockListFormatter.read_packages_from_lock_file(f)

        session = requests.Session()

        packages_not_found = []

        output_path = self._config.output_path

        if output_path is not None and output_path != '':
            pathlib.Path(output_path).mkdir(parents=True, exist_ok=True)

        for package_path in packages_path_in_repo:

            errors = {}
            package_download_done = False

            for src in self._config.sources():
                if package_download_done:
                    break

                session.auth = src.get_auth_params().auth
                packages = src.generate_full_urls_from_package_path_in_repo(package_path)

                # packages = self.find_package_in_artifactory(src, package)
                for p in packages:
                    try:
                        ap = ArtifactoryPath(p, session=session)
                        dst_path = os.path.join(output_path, ap.name)

                        with ap.open() as input, open(dst_path, "wb") as output:
                            shutil.copyfileobj(input, output)

                        package_download_done = True

                        self._log.info(f"Success {ap} downloaded to {dst_path}")

                        break

                    except RuntimeError as e:
                        if isinstance(e.args[0], int):
                            errors[ap] = http.HTTPStatus(e.args[0])
                        else:
                            errors[ap] = e

            if not package_download_done:
                packages_not_found += [package_path]
                self._log.error(f"PACKAGE NOT FOUND {package_path}. Errors for attempts {errors}")

        if packages_not_found:
            raise CrosspmException(CROSSPM_ERRORCODE_PACKAGE_NOT_FOUND,
                                   'Some package(s) not found: {}'.format(', '.join(packages_not_found)))

    def entrypoint(self, *args, **kwargs):
        self.download_packages(*args, **kwargs)

    def search_dependencies(self, depslock_file_path, deps_content=None):
        self._log.info('Check dependencies ...')
        self._root_package.find_dependencies(depslock_file_path, property_validate=True, deps_content=deps_content, )
        self._log.info('')
        self.check_not_found()

    def check_not_found(self):
        _not_found = self.get_not_found_packages()
        if _not_found:
            raise CrosspmException(
                CROSSPM_ERRORCODE_PACKAGE_NOT_FOUND,
                'Some package(s) not found: {}'.format(', '.join(_not_found))
            )

    def get_not_found_packages(self):
        return self._root_package.get_none_packages()

    def add_package(self, pkg_name, package):
        _added = False
        if package is not None:
            _added = True
        return _added, package

    def get_raw_packages(self):
        """
        Get all packages
        :return: list of all packages
        """
        return self._root_package.all_packages

    def get_tree_packages(self):
        """
        Get all packages, with hierarchy
        :return: list of first level packages, with child
        """
        return self._root_package.packages

    def iter_packages_params(self, list_or_file_path, deps_content=None):
        return self.common_parser.iter_packages_params(list_or_file_path, deps_content)



# -*- coding: utf-8 -*-
from collections import OrderedDict

import json
import packaging
import requests
from addict import Dict
from artifactory import ArtifactoryPath

import crosspm.contracts.package
from crosspm import InvalidPackage
from crosspm.adapters.common import BaseAdapter
from crosspm.contracts.bundle import Bundle
from crosspm.helpers.exceptions import *  # noqa
from crosspm.helpers.package import Package
from dohq_common.exceptions import PackageInvalidVersion

CHUNK_SIZE = 1024

setup = {
    "name": [
        "jfrog-artifactory-aql2",
    ],
}

session = requests.Session()


class ArtifactoryAql2(BaseAdapter):
    def get_packages(self, source, parser, downloader, packages_matches, property_validate=True):
        # TODO move request debug and thhp debug into comanline option
        # http.client.HTTPConnection.debuglevel = 1
        # requests_log = logging.getLogger("requests.packages.urllib3")
        # requests_log.setLevel(logging.DEBUG)
        # requests_log.propagate = True

        _art_auth_etc = source.get_auth_params()

        _packages_found = OrderedDict()
        self._log.info('parser: {}'.format(parser._name))

        repo_returned_packages_all = []

        for _path in source.get_paths(packages_matches):

            _tmp_params = Dict(_path.params)
            self._log.info('repo: {}'.format(_tmp_params.repo))

            session.auth = _art_auth_etc['auth']
            aql = ArtifactoryPath(f"{_tmp_params.server}", session=session)

            _path_fixed, _path_pattern, _file_name_pattern = parser.split_fixed_pattern_with_file_name(_path.path)

            _package_versions_with_contracts, packages_with_invalid_naming_convention = \
                self.find_package_versions(_file_name_pattern,
                                           _path_pattern,
                                           aql,
                                           _tmp_params.repo)

            if _package_versions_with_contracts:
                _package_versions_with_all_contracts, package_versions_with_missing_contracts = \
                    remove_package_versions_with_missing_contracts(
                        _package_versions_with_contracts,
                        _path.params['contracts']
                    )

                for p, missing_contracts in package_versions_with_missing_contracts.items():
                    self._log.info(f"Skip {p} - missing contracts {missing_contracts}")

                if _package_versions_with_all_contracts:
                    repo_returned_packages_all += _package_versions_with_all_contracts

        package_names = [x.package_name for x in packages_matches]

        bundle = Bundle(package_names, repo_returned_packages_all,
                        downloader._config.trigger_packages, enable_tp_hides_higher_version=False)

        bundle_packages = bundle.calculate().values()

        self._log.info('Final bundle packages with contracts:')
        print_packages_by_contracts_scheme(self._log, bundle_packages)

        for p in bundle_packages:
            _packages_found[p.name] = Package(p.name, p.art_path, p, {}, downloader, self, parser,
                                              {}, {})

        for p in package_names:
            if p not in _packages_found.keys():
                _packages_found[p.name] = None

        return _packages_found

    def find_package_versions(self, _file_name_pattern,
                              _path_pattern, aql, search_repo):
        try:
            packages_with_invalid_naming_convention = []
            package_versions_with_contracts = []

            query = self.prepare_aql_query(_file_name_pattern, _path_pattern, search_repo)
            artifacts = aql.aql(query)

            for art_path in map(aql.from_aql, artifacts):
                try:

                    package_with_contracts = crosspm.contracts.package.create_artifactory_package(art_path)

                    package_versions_with_contracts.append(package_with_contracts)

                    self._log.debug(f"  valid: {str(art_path)}")

                except PackageInvalidVersion:
                    pass
                except packaging.version.InvalidVersion as e:
                    packages_with_invalid_naming_convention.append(InvalidPackage(art_path, e))
                    self._log.warn(f"{e} for {art_path}")

        except RuntimeError as e:
            self.try_parse_http_error(e)

        return package_versions_with_contracts, packages_with_invalid_naming_convention

    def prepare_aql_query(self, _file_name_pattern, _path_pattern, _search_repo):
        # Get AQL path pattern, with fixed part path, without artifactory url and repository name
        _aql_path_pattern = ""
        if _path_pattern:
            _aql_path_pattern = _path_pattern
        _aql_query_dict = {
            "repo": {
                "$eq": _search_repo,
            },
            "path": {
                "$match": _aql_path_pattern,
            },
            "name": {
                "$match": _file_name_pattern,
            },
        }
        # Remove path if is empty string
        if not _aql_path_pattern:
            _aql_query_dict.pop('path')
        query = f'items.find({json.dumps(_aql_query_dict)}).include("*", "property")'

        return query

    def try_parse_http_error(self, e):
        try:
            err = json.loads(e.args[0])
        except Exception:
            err = {}
        if isinstance(err, dict):
            # Check errors
            # :e.args[0]: {
            #                 "errors" : [ {
            #                     "status" : 404,
            #                     "message" : "Not Found"
            #                  } ]
            #             }
            for error in err.get('errors', []):
                err_status = error.get('status', -1)
                err_msg = error.get('message', '')

                self._log.error('Error[{}]{}'.format(err_status,
                                                     (': {}'.format(err_msg)) if err_msg else ''))


def print_packages_by_contracts_scheme(logger, packages):
    for p in packages:
        logger.info(f"  {p}")


def remove_package_versions_with_missing_contracts(package_versions, contracts):
    if not contracts:
        return package_versions, {}

    package_versions_with_all_contracts = []
    package_versions_with_missing_contracts = {}

    for p in package_versions:
        package_version_missing_contracts = []
        for c in contracts:
            if not p.has_contract(c):
                package_version_missing_contracts += [c]

        if package_version_missing_contracts:
            package_versions_with_missing_contracts[p] = package_version_missing_contracts
        else:
            package_versions_with_all_contracts += [p]

    return package_versions_with_all_contracts, package_versions_with_missing_contracts
